{----------------------------------------------------------------------------
|
| Library: Spider Containers for Object Pascal
|
| Module: ObjBTree.Pas
|
| Description: Binary tree container.
|
| Copyright (c) 1996,97 Michel Brazeau, Interval Software
|
| History: Version 1.0  March 1996. 
|
|---------------------------------------------------------------------------}
unit ObjBTree;

interface

{$I Contain.Inc}

uses
    ObjPerst, { TObjectReader, TObjectWriter }
    ObjCmp,   { TCompareFunc }
    ObjBase;  { TNodeContainer }

type

{ binary tree node }
TBTreeNode = class(TNode)
private
    function GetSize : LongInt;

    function GetOwnership : Boolean;
    procedure SetOwnership( const InOwnership : Boolean );

public
    Parent,
    Left,
    Right          : TBTreeNode;

    { the base class Obj member is created as a TUnOrderedList to allow
      duplicate items }

    constructor Create(const InOwnership : Boolean);

    { stores an object in the node }
    procedure StoreObj(InObj : TObject);

    { detaches an object from the list, and returns the object. }
    function DetachNthObj(Index : LongInt) : TObject;

    { Gets the N'th object from the node. }
    function GetNthObj(Index : LongInt) : TObject;

    { returns the first object in the node }
    function GetFirstObj : TObject;

    { returns the number of objects in the node }
    property Size : LongInt read GetSize;

    { if True the objects in the node are freed when the node is freed. }
    property Ownership : Boolean read GetOwnership
                                 write SetOwnership;

end;

type

{ Binary tree class }
TBinaryTree = class(TNodeContainer)
public
    constructor Create( InObjectType   : TClass;
                        InCompareFunc  : TCompareFunc
                       ); 

    destructor Destroy; override;

    { inserts a node at the proper position according to the compare
      function }
    procedure Insert(const Obj : TObject); override;

    { Deletes (removes and Frees) the N'th matching Occur'ence of Obj.
      Returns True if the object is found and deleted. Positions the
      current node to the next node if possible }
    function Delete( const Obj   : TObject;
                     const Occur : LongInt
                    ) : Boolean; override;

    { deletes the current node, without freeing the object in the node.
      It is then the responsibility of the user to retieve the object first,
      in order to later free the object. The parent is then made the
      Current node. May raise EContainerIsEmpty }
    function DetachCurrent : TObject; override;

    { searches for the N'th matching Occur'ence of Obj, and if found, the
      current node is set to it. }
    function Search( const Obj   : TObject;
                     const Occur : LongInt
                    ) : Boolean; override;

    { GotoFirst, GotoLast return False when the tree is emtpy.
      EContainerIsEmpty is not raised. }
    function GotoFirst : Boolean; override;
    function GotoLast  : Boolean; override;

    { for each item in the container, call ProcIterator. ProcIterator is
      a file level procedure. }
    procedure ForEachCallProc( const ProcIterator : TProcIterator;
                               const CallBackData   : LongInt); override;

    { for each item in the container, call MethodIterator. MethodIterator is
      a method of a class. }
    procedure ForEachCallMethod( const MethodIterator : TMethodIterator;
                                 const CallBackData   : LongInt); override;

    { returns False if list is empty or last node is reached }
    function GotoNext : Boolean; override;

    { returns False if list is empty or first node is reached }
    function GotoPrev : Boolean; override;

    procedure Clear; override;

protected

    Root     : TBtreeNode;

    CurOccur : LongInt;  { last occurence processed within the current node }

    { returns True if there is a node that contains one or more duplicates
      of Obj and Current points to the node if found. If not found, Current
      points to the position where the node would be inserted }
    function SearchNode(const Obj  : TObject) : Boolean;

    function  GetCurrentObj : TObject; override;

    function GetNthObj(const Index : LongInt) : TObject; override;

    { positions current to the first or last record from the current node }
    function GotoFirstFromCurrent : Boolean;
    function GotoLastFromCurrent  : Boolean;

end;

{ stream registration routines }
procedure ReadBinaryTree( const Obj    : TObject;
                          const Reader : TObjectReader ); far;

procedure WriteBinaryTree( const Obj    : TObject;
                           const Writer : TObjectWriter ); far;

{--------------------------------------------------------------------}

implementation

uses
    ObjType,  { RaiseXXX }
    ObjList;  { TUnOrderedList }

{ TBTreeNode implemenation -----------------------------------------------}

constructor TBTreeNode.Create(const InOwnership : Boolean);
begin
    inherited Create;

    Obj := TUnOrderedList.Create(TObject, CompareEqual);

    TUnOrderedList(Obj).Ownership := InOwnership;
end;

{-------------------------------------------------------------------------}

function TBTreeNode.GetSize : LongInt;
begin
    Result := (Obj as TUnOrderedList).Size;
end;

{-------------------------------------------------------------------------}

procedure TBTreeNode.StoreObj(InObj : TObject);
begin
    (Obj as TUnOrderedList).InsertAtTail(InObj);
end;

{-------------------------------------------------------------------------}

function TBTreeNode.GetNthObj(Index : LongInt) : TObject;
begin
    Result := (Obj as TUnOrderedList)[Index];
end;

{-------------------------------------------------------------------------}

function TBTreeNode.DetachNthObj(Index : LongInt) : TObject;
begin
    { get the object }
    Result := (Obj as TUnOrderedList)[Index];

    { detach the node }
    (Obj as TUnOrderedList).DetachCurrent;
end;

{-------------------------------------------------------------------------}

function TBTreeNode.GetFirstObj : TObject;
begin
    Result := nil;

    if (Obj as TUnOrderedList).GotoFirst then
        Result := (Obj as TUnOrderedList).CurrentObj;
end;

{-------------------------------------------------------------------------}

function TBTreeNode.GetOwnership : Boolean;
begin
    Result := (Obj as TUnOrderedList).Ownership;
end;

{-------------------------------------------------------------------------}

procedure TBTreeNode.SetOwnership( const InOwnership : Boolean );
begin
    (Obj as TUnOrderedList).Ownership := InOwnership;
end;

{ TBinaryTree methods ----------------------------------------------------}

constructor TBinaryTree.Create( InObjectType   : TClass;
                                InCompareFunc  : TCompareFunc
                               );
begin
    inherited Create(InObjectType, InCompareFunc);

    CurOccur := 1;

    Root     := nil;
end;

{-------------------------------------------------------------------------}

destructor TBinaryTree.Destroy;
begin
    inherited Destroy;
end;

{-------------------------------------------------------------------------}

procedure TBinaryTree.Insert(const Obj : TObject);
var
    Node : TBTreeNode;

begin
    {$ifndef __RemoveTypeSafety}
    CheckType(Obj);
    {$endif}

    { check for duplicate }
    if Search(Obj, 1) then
    begin
        (Current as TBTreeNode).StoreObj(Obj);  { add object to node }
        CurOccur := (Current as TBTreeNode).Size;
        Inc(FSize);
        Exit;
    end;

    { assert: insertion position was set in Search }

    Node := TBTreeNode.Create(Self.Ownership);
    Node.StoreObj(Obj);
    Node.Right  := nil;
    Node.Left   := nil;
    Node.Parent := (Current as TBTreeNode);

    if Current <> nil then
    begin
        if FCompareFunc((Current as TBTreeNode).GetFirstObj, Obj) = cmpGreater then
            (Current as TBTreeNode).Left := Node
        else
            (Current as TBTreeNode).Right := Node
    end
    else
        Root := Node;

    CurOccur := 1;

    Inc(FSize);
end;


{-------------------------------------------------------------------------}

function TBinaryTree.Delete( const Obj   : TObject;
                             const Occur : LongInt
                            ) : Boolean;
begin
    {$ifndef __RemoveTypeSafety}
    CheckType(Obj);
    {$endif}

    Result := Search(Obj, Occur);

    if Result then
        { assert : Current and CurOccur point to the found record }
        DeleteCurrent;
end;

{-------------------------------------------------------------------------}

function TBinaryTree.DetachCurrent : TObject;
var
    Node : TBTreeNode;
begin
    if Current = nil then
        RaiseContainerIsEmpty;

    if (Current as TBTreeNode).Size > 1 then
    begin
        { don't remove the node, just the duplicate }
        Result := (Current as TBTreeNode).DetachNthObj(CurOccur);

        Dec(FSize);

        if CurOccur > (Current as TBTreeNode).Size then
            Dec(CurOccur);

        Exit;
    end;

    Node := (Current as TBTreeNode);

    if Node.Right <> nil then
    begin { replace the Current node with the right branch }
        Current      := Node.Right;

        GotoFirstFromCurrent;

        (Current as TBTreeNode).Left := Node.Left;
        Node.Right.Parent            := Node.Parent;

        if Node.Left <> nil then
            Node.Left.Parent := (Current as TBTreeNode);

        if Node.Parent <> nil then
        begin
            if Node.Parent.Left = Node then
                Node.Parent.Left := Node.Right
            else
                Node.Parent.Right := Node.Right
        end
        else
            Root := Node.Right;
    end
    else if Node.Left <> nil then
    begin { replace the Current node with the left branch }
        Node.Left.Parent := Node.Parent;

        if Node.Parent <> nil then
            if Node.Parent.Right = Node then
                Node.Parent.Right := Node.Left
            else
                Node.Parent.Left := Node.Left
        else
            Root := Node.Left;
    end
    else if Node.Parent <> nil then
    begin
        { Current node is a leaf }
        if Node.Parent.Right = Node then
            Node.Parent.Right := nil
        else
            Node.Parent.Left := nil
    end
    else
    begin
        { current node is only node in tree }
        Root    := nil;
        Current := nil;
    end;

    { position current as the parent of the deleted node }
    if Root <> nil then
        Current := Node.Parent;

    Result := Node.DetachNthObj(1);

    Node.Free;

    CurOccur := 1; { reset to first occurence of current node }

    Dec(FSize);

end;

{-------------------------------------------------------------------------}

function TBinaryTree.Search( const Obj     : TObject;
                             const Occur   : LongInt
                           ) : Boolean;
begin
    {$ifndef __RemoveTypeSafety}
    CheckType(Obj);
    {$endif}

    Result := False;

    if Occur <= 0 then
        RaiseInvalidOccurrenceSought;

    if SearchNode(Obj) then
    begin
        if (Current as TBTreeNode).Size >= Occur then
        begin
            CurOccur := Occur;
            Result   := True;
        end
        else
            CurOccur := 1;
    end;
end;

{-------------------------------------------------------------------------}

function TBinaryTree.SearchNode( const Obj : TObject
                                ) : Boolean;
begin
    { this is a protected method, so CheckType has already been called }
    Result := False;

    Current := Root;

    while Current <> nil do
    begin
        if FCompareFunc((Current as TBTreeNode).GetFirstObj, Obj) = cmpEqual then
        begin
            Result := True;
            Exit;
        end
        else if FCompareFunc((Current as TBTreeNode).GetFirstObj, Obj) = cmpGreater then
        begin
            if (Current as TBTreeNode).Left <> nil then
                Current := (Current as TBTreeNode).Left
            else
                Exit
        end
        else if (Current as TBTreeNode).Right <> nil then
            Current := (Current as TBTreeNode).Right
        else
            Exit;
    end;
end;

{-------------------------------------------------------------------------}

function  TBinaryTree.GetCurrentObj : TObject;
begin
    if Current = nil then
        RaiseContainerIsEmpty;

    Result := (Current as TBtreeNode).GetNthObj(CurOccur);
end;

{-------------------------------------------------------------------------}

function TBinaryTree.GetNthObj(const Index : LongInt) : TObject;
var
    I     : LongInt;

begin
    { check bounds on N }
    if (Index<1) or (Index>Size) then
        RaiseInvalidIndex;

    if Index < (Size div 2) then
    begin
	GotoFirst;

	for I := 2 to Index do
	    GotoNext;
    end
    else
    begin
	GotoLast;

	for I := 1 to (Size-Index) do
	    GotoPrev;
    end;

    Result := CurrentObj;
end;

{-------------------------------------------------------------------------}

function TBinaryTree.GotoFirstFromCurrent : Boolean;
begin
    Result := (Current <> nil);

    if Current <> nil then
    begin
        while (Current as TBTreeNode).Left <> nil do
            Current := (Current as TBTreeNode).Left;

        CurOccur := 1;
    end;

end;

{-------------------------------------------------------------------------}

function TBinaryTree.GotoLastFromCurrent : Boolean;
begin
    Result := (Current <> nil);

    if Current <> nil then
    begin
        while (Current as TBTreeNode).Right <> nil do
            Current := (Current as TBTreeNode).Right;

        CurOccur := (Current as TBTreeNode).Size;
    end;

end;

{-------------------------------------------------------------------------}

function TBinaryTree.GotoFirst : Boolean;
begin
    Current  := Root;
    CurOccur := 1;

    Result := GotoFirstFromCurrent;
end;

{-------------------------------------------------------------------------}

function TBinaryTree.GotoLast  : Boolean;
begin
    Current  := Root;
    CurOccur := (Current as TBTreeNode).Size;

    Result := GotoLastFromCurrent;
end;

{-------------------------------------------------------------------------}

function TBinaryTree.GotoNext : Boolean;
var
    SavedCurrent : TBTreeNode;
begin
    Result := True;

    if Current = nil then
    begin
        Result := False;
        Exit;
    end;

    if (Current as TBTreeNode).Size > CurOccur then
    begin
        { move next within the current node }
        Inc(CurOccur);
        Exit;
    end;

    SavedCurrent := (Current as TBTreeNode);

    if (Current as TBTreeNode).Right <> nil then
    begin
        Current := (Current as TBTreeNode).Right;
        while (Current as TBTreeNode).Left <> nil do
            Current := (Current as TBTreeNode).Left;
    end
    else
    begin
        while ((Current as TBTreeNode).Parent <> nil) and
              ((Current as TBTreeNode).Parent.Right = Current) do
            Current := (Current as TBTreeNode).Parent;

        if (Current as TBTreeNode).Parent <> nil then
            Current := (Current as TBTreeNode).Parent
        else
        begin
            Current := SavedCurrent;
            Result  := False; { Current was initially at the largest node }
        end;
    end;

    if Result then
        CurOccur := 1;  { move to first object of new node }
end;

{-------------------------------------------------------------------------}

function TBinaryTree.GotoPrev : Boolean;
var
    SavedCurrent : TBTreeNode;
begin
    Result := True;

    if Current = nil then
    begin
        Result := False;
        Exit;
    end;

    if CurOccur > 1 then
    begin
        { move previous within the current node }
        Dec(CurOccur);
        Exit;
    end;

    SavedCurrent := (Current as TBTreeNode);

    if (Current as TBTreeNode).Left <> nil then
    begin
        Current := (Current as TBTreeNode).Left;
        while (Current as TBTreeNode).Right <> nil do
            Current := (Current as TBTreeNode).Right;
    end
    else
    begin
        while ((Current as TBTreeNode).Parent <> nil) and
              ((Current as TBTreeNode).Parent.Left = Current) do
            Current := (Current as TBTreeNode).Parent;

        if (Current as TBTreeNode).Parent <> nil then
            Current := (Current as TBTreeNode).Parent
        else
        begin
            Current := SavedCurrent;
            Result  := False;
        end;
    end;

    if Result then
        CurOccur := (Current as TBTreeNode).Size;  { move to last object of new node }

end;

{-----------------------------------------------------------------------}

procedure TBinaryTree.ForEachCallProc( const ProcIterator : TProcIterator;
                                       const CallBackData   : LongInt);
begin
    if GotoFirst then
    repeat
        ProcIterator(CurrentObj, CallBackData);
    until not GotoNext;
end;

{-----------------------------------------------------------------------}

procedure TBinaryTree.ForEachCallMethod( const MethodIterator : TMethodIterator;
                                         const CallBackData   : LongInt);
begin
    if GotoFirst then
    repeat
        MethodIterator(CurrentObj, CallBackData);
    until not GotoNext;
end;

{-------------------------------------------------------------------------}

procedure TBinaryTree.Clear;
begin
    { remove the root node, until the list is empty }
    while Size > 0 do
    begin
        CurOccur := 1;     { to ensure entire node is removed }
        Current  := Root;

        DeleteCurrent;
    end;

    CurOccur := 1;
    Current  := nil;
    Root     := nil;
end;

{-------------------------------------------------------------------------}

procedure ReadBinaryTree( const Obj    : TObject;
                          const Reader : TObjectReader );
begin
    with Obj as TBinaryTree do
    begin
        CurOccur := 1;
        Root     := nil;
    end;

    ReadNodeContainer( Obj, Reader );
end;

{-------------------------------------------------------------------------}

procedure WriteBinaryTree( const Obj    : TObject;
                           const Writer : TObjectWriter );
begin
    WriteNodeContainer( Obj, Writer );
end;

{---------------------------------------------------------------------}

initialization

    {$ifdef Win32}
    RegisterStreamable( TBinaryTree,
                        [TObjectReadProc(@ReadBinaryTree)],
                        [TObjectWriteProc(@WriteBinaryTree)] );
    {$else}
    RegisterStreamable( TBinaryTree,
                        [ReadBinaryTree],
                        [WriteBinaryTree] );
    {$endif}

end.
