{----------------------------------------------------------------------------
|
| Library: Spider Object Database
|
| Module: OdbBtr.Pas
|
| Description: Btrieve object database driver class.
|
|              This driver supports Btrieve 6.15, Pervasive.SQL 7 and 2000.
|              Future versions should also be supported as Pervasive software
|              has a strong commitment to the Btrieve API
|
| History: Jan 25, 1998. Michel Brazeau, first version
|          Mar 22, 2001. MB updated for version.
|
|---------------------------------------------------------------------------}
unit OdbBtr;

{$I Odb.Inc}

interface

uses
    {$ifdef __Testing}
    DriverTest, { for TOdbDriverTest }
    {$endif}

    Classes,  { for TMemoryStream }
    SysUtils, { for Exception }
    ObjList,  { for TUnOrderedList }
    Odb,      { for TObjectFile, TOdbDriver }
    OdbIndex, { for TOdbIndex }
    OdbField, { for TOdbField }
    OdbBtrFl; { for TBtrieveFile }

type

{--------------------------------------------------------------------------}

TBtrieveOdbDriver = class(TOdbDriver)
protected

    FDataBuffer   : array[1..80] of Byte;
    FPositionRec  : TBtPositionRec;
    FKeyBuffer    : array[1..256] of Byte;

    FClientId     : TBtClientId;

    FChunkSize    : Word;

    procedure SetChunkSize( const ChunkSize : Word );

public

    constructor Create(const InDatabaseName : String); override;
    destructor Destroy; override;

    { Returns a pointer to the client ID parameter used with the Btrieve API
      call. }
    function GetClientIdPtr : TpBtClientId;

    { Maximum size which will be used with Btrieve chunk operations. Range is
      from 1024 to 51200. A large value improve performance when storing
      large objects, but will require larger setup paramaters with the Btrieve
      MicroKernel engine. Default is 12288. }
    property ChunkSize : Word read FChunkSize
                              write SetChunkSize;

    procedure InternalStartTransaction; override;
    procedure InternalCommit; override;
    procedure InternalRollback; override;

    procedure InternalOpenDatabase(
                            const ServerName : String;
                            const UserName   : String;
                            const Password   : String
                          ); override;

    procedure InternalCloseDatabase; override;

    function InternalOpenDataset( const ObjectDataset : TObjectDataset ) : TOdbHandle; override;

    function InternalOpenExclusive( const ObjectFile : TObjectFile ) : TOdbHandle; override;

    procedure InternalCloseDataset( const Handle : TOdbHandle); override;

    procedure InternalDeleteDataset( const ObjectDataset : TObjectDataset ); override;

    procedure InternalCreateDataset( const ObjectDataset : TObjectDataset ); override;

    procedure InternalInsert(
                      const Handle  : TOdbHandle;
                      const Fields  : TUnOrderedList;
                      const Stream  : TMemoryStream ); override;

    procedure InternalUpdate(
                      const Handle     : TOdbHandle;
                      const Fields     : TUnOrderedList;
                      const Stream     : TMemoryStream ); override;

    procedure InternalDelete( const Handle : TOdbHandle ); override;

    function InternalGetIndexName( const Handle : TOdbHandle ) : String; override;

    procedure InternalSetIndexName(
                            const Handle    : TOdbHandle;
                            const IndexName : String ); override;

    procedure InternalGetCurrent(
                          const Handle  : TOdbHandle;
                          const Stream  : TMemoryStream ); override;

    function InternalLockCurrent( const Handle : TOdbHandle ) : Boolean; override;

    procedure InternalUnLockCurrent( const Handle : TOdbHandle ); override;

    function InternalFindNext( const Handle : TOdbHandle ) : Boolean; override;

    function InternalFindPrior( const Handle : TOdbHandle ) : Boolean; override;

    function InternalFindFirst( const Handle : TOdbHandle ) : Boolean; override;

    function InternalFindLast( const Handle : TOdbHandle ) : Boolean; override;

    function InternalFindExact(
                        const Handle : TOdbHandle;
                        const Fields : TUnOrderedList;
                        const Index  : TOdbIndex ) : Boolean; override;

    procedure InternalFindNearest(
                          const Handle : TOdbHandle;
                          const Fields : TUnOrderedList;
                          const Index  : TOdbIndex ); override;

    function InternalGetBookmark( const Handle : TOdbHandle ) : Pointer; override;
    procedure InternalGotoBookmark( const Handle   : TOdbHandle;
                                    const Bookmark : Pointer ); override;
    procedure InternalFreeBookmark( const Handle   : TOdbHandle;
                                    const Bookmark : Pointer ); override;

    function InternalGetObjectCount( const Handle : TOdbHandle ) : LongInt; override;

    function GetSupportedFieldTypes : TOdbFieldTypes; override;

    function GetBtrieveFile( const Handle : TOdbHandle ) : TBtrieveFile;

    { returns the version for Btrieve loaded which is responsible to
      access files located in directory defined by the database name.
      The version is returned in a string using the following format:

      Version.Revision.Engine, for example
      6.15N netware requester
      6.15D dos
      6.15P OS/2
      6.15W windows
      6.15S netware server.
      6.15T NT server

      If the requester and server is loaded the version of the server
      is returned. If the requested is loaded, but the server is not
      available (or cannot access files on directory) then the version of
      the requester is returned.

      If the requester, or the client version of Btrieve is not active,
      then the EOdbBtrieveError is raised with the status code.
    }
    function GetBtrieveVersion : String;

    function GetDataAccessMethods : TDataAccessMethodSet; override;

    function AutoIncSupported              : Boolean; override;
    function IntegerIndexSupported         : Boolean; override;
    function FloatIndexSupported           : Boolean; override;
    function CaseInsensitiveIndexSupported : Boolean; override;
    function DescendingIndexSupported      : Boolean; override;
    function TransactionsSupported         : Boolean; override;
    function FileLockingSupported          : Boolean; override;
    function ObjectLockingSupported        : Boolean; override;
    function SingleIndexOnly               : Boolean; override;

protected
    function BtrieveOpenFile( const Name       : String;
                              const ObjectFile : TObjectFile;
                              const Exclusive  : Boolean ) : TOdbHandle;



end;


{$ifdef __Testing}

type

TBtrieveOdbDriverTest = class( TOdbDriverTest )
protected
     procedure GetParams( var DriverClass                   : TOdbDriverClass;
                          var DatabaseName                  : String;
                          var ServerName                    : String;
                          var UserName                      : String;
                          var Password                      : String
                        ); override;
end;

{$endif}

{--------------------------------------------------------------------}

implementation

uses
    FileCtrl, { for DirectoryExists }
    OdbMisc;  { for RaiseOdbException }

const

CMinChunkSize     = 1024;
CMaxChunkSize     = 51200;
CDefaultChunkSize = 12288;

{--------------------------------------------------------------------------}

constructor TBtrieveOdbDriver.Create(const InDatabaseName : String);
begin
    if (InDatabaseName <> '') and
       not DirectoryExists(InDatabaseName) then
        RaiseOdbException( EOdbDatabaseDoesNotExist, InDatabaseName);

    NewClientId(FClientId);

    FChunkSize := CDefaultChunkSize;

    inherited Create(InDatabaseName);
end;

{--------------------------------------------------------------------------}

destructor TBtrieveOdbDriver.Destroy;
begin
    inherited Destroy;
end;

{--------------------------------------------------------------------------}

function TBtrieveOdbDriver.GetClientIdPtr : TpBtClientId;
begin
    Result := @FClientId;
end;

{--------------------------------------------------------------------------}

procedure TBtrieveOdbDriver.SetChunkSize( const ChunkSize : Word );
begin
    if (FChunkSize <> ChunkSize) then
    begin
        FChunkSize := ChunkSize;

        if FChunkSize < CMinChunkSize then
            FChunkSize := CMinChunkSize;

        if FChunkSize > CMaxChunkSize then
            FChunkSize := CMaxChunkSize;
    end;

end;

{--------------------------------------------------------------------------}

procedure TBtrieveOdbDriver.InternalStartTransaction;
var
    DataLen   : Word;
begin
    DataLen := 0;
    BtrieveCheck(BtrvId( bBeginTransaction+bConCurrentTransBias+bMultiNoWaitLock,
                         FPositionRec, FDataBuffer,
                         DataLen,FKeyBuffer, 0, FClientId), '');
end;

{--------------------------------------------------------------------------}

procedure TBtrieveOdbDriver.InternalCommit;
var
    DataLen   : Word;
begin
    DataLen := 0;
    BtrieveCheck( BtrvId( bEndTransaction,
                          FPositionRec, FDataBuffer,
                          DataLen, FKeyBuffer, 0, FClientId), '');
end;

{--------------------------------------------------------------------------}

procedure TBtrieveOdbDriver.InternalRollback;
var
    DataLen   : Word;
begin
    DataLen := 0;
    BtrieveCheck( BtrvId( bAbortTransaction,
                          FPositionRec, FDataBuffer,
                          DataLen, FKeyBuffer, 0, FClientId), '');
end;

{--------------------------------------------------------------------------}

function TBtrieveOdbDriver.GetBtrieveVersion : String;
const
    CVersionLen = 5; { number of bytes to specify btrieve version. Specific
                       to the Btrieve API }
var
    DataLen       : Word;
    TempDir       : String[254]; { 254 to allow null terminating FKeyBuffer }
    Version       : Word;
    Revision      : Word;
    VersionOffset : Word;

begin
    DataLen := 3 * CVersionLen; { specific to the bVersion call }

    TempDir := DatabaseName;
    TempDir := RemoveBackSlashFromPath(TempDir);

    Move(TempDir[1], FKeyBuffer[1], Length(TempDir));

    { null terminate the directory name }
    FKeyBuffer[Length(TempDir)+1] := 0;

    BtrieveCheck( BtrvId( bVersion,
                          FPositionRec, FDataBuffer,
                          DataLen, FKeyBuffer, 0, FClientId), '');

    VersionOffset := DataLen div CVersionLen - 1;
    Version  := Word( Pointer(@FDataBuffer[VersionOffset*CVersionLen + 1])^ );
    Revision := Word( Pointer(@FDataBuffer[VersionOffset*CVersionLen + 3])^ );

    Result    := IntToStr(Version) +
                 '.' +
                 IntToStr(Revision) +
                 Char( FDataBuffer[VersionOffset*CVersionLen+ 5] );
end;

{--------------------------------------------------------------------------}

procedure TBtrieveOdbDriver.InternalOpenDatabase(
                        const ServerName : String;
                        const UserName   : String;
                        const Password   : String
                      );
begin
    { attempt to get the Btrieve version. This will raise an exception if
      the Btrieve engine is not installed. }
    GetBtrieveVersion;
end;

{--------------------------------------------------------------------------}

procedure TBtrieveOdbDriver.InternalCloseDatabase;
var
    DataLen : Word;

begin
    DataLen := 0;
    FillChar(FKeyBuffer, SizeOf(FKeyBuffer), 0);

    BtrieveCheck( BtrvId( bStop,
                          FPositionRec, FDataBuffer,
                          DataLen, FKeyBuffer, 0, FClientId), '');
end;

{--------------------------------------------------------------------------}

function TBtrieveOdbDriver.BtrieveOpenFile(
                                    const Name       : String;
                                    const ObjectFile : TObjectFile;
                                    const Exclusive  : Boolean ) : TOdbHandle;
var
    BtrieveFile : TBtrieveFile;

begin
    BtrieveFile := TBtrieveFile.Create(ObjectFile);
    try
        if Exclusive then
            BtrieveFile.OpenExclusive
        else
            BtrieveFile.Open;

        Result := NewHandle(BtrieveFile);
    except
        BtrieveFile.Free;
        raise;
    end;
end;

{--------------------------------------------------------------------------}

function TBtrieveOdbDriver.InternalOpenDataset(
                       const ObjectDataset : TObjectDataset ) : TOdbHandle;
begin
    Result := BtrieveOpenFile(ObjectDataset.DatasetName, ObjectDataset as TObjectFile, False);
end;
{--------------------------------------------------------------------------}

function TBtrieveOdbDriver.InternalOpenExclusive(
                                  const ObjectFile : TObjectFile ) : TOdbHandle;
begin
    Result := BtrieveOpenFile(ObjectFile.DatasetName, ObjectFile, True);
end;

{--------------------------------------------------------------------------}

procedure TBtrieveOdbDriver.InternalCloseDataset( const Handle : TOdbHandle);
begin
    GetBtrieveFile(Handle).Close;
    ReleaseHandle(Handle);
end;

{--------------------------------------------------------------------------}

procedure TBtrieveOdbDriver.InternalDeleteDataset( const ObjectDataset : TObjectDataset );
var
    FileName    : String;
begin
    FileName := GetLocalObjectFilename( DatabaseName,
                                        ObjectDataset.DatasetName,
                                        CBtrieveFileExtension);
    SysUtils.DeleteFile(FileName);
end;

{--------------------------------------------------------------------------}

procedure TBtrieveOdbDriver.InternalCreateDataset(
                          const ObjectDataset : TObjectDataset
                        );
var
    BtrieveFile : TBtrieveFile;

begin
    BtrieveFile := TBtrieveFile.Create(ObjectDataset as TObjectFile);
    try
        BtrieveFile.CreateFile;
    finally
        BtrieveFile.Free;
    end;
end;

{--------------------------------------------------------------------------}

procedure TBtrieveOdbDriver.InternalInsert(
                  const Handle  : TOdbHandle;
                  const Fields  : TUnOrderedList;
                  const Stream  : TMemoryStream );
var
    BtrieveFile   : TBtrieveFile;
begin
    BtrieveFile := GetBtrieveFile(Handle);
    BtrieveFile.Insert( Fields, Stream );
end;

{--------------------------------------------------------------------------}

procedure TBtrieveOdbDriver.InternalUpdate(
                  const Handle     : TOdbHandle;
                  const Fields     : TUnOrderedList;
                  const Stream     : TMemoryStream );
var
    BtrieveFile : TBtrieveFile;
begin
    BtrieveFile := GetBtrieveFile(Handle);
    BtrieveFile.Update( Fields, Stream );
end;

{--------------------------------------------------------------------------}

procedure TBtrieveOdbDriver.InternalDelete( const Handle : TOdbHandle );
var
    BtrieveFile : TBtrieveFile;
begin
    BtrieveFile := GetBtrieveFile(Handle);
    BtrieveFile.Delete;
    if not BtrieveFile.Next then
        BtrieveFile.Prior;
end;

{--------------------------------------------------------------------------}

function TBtrieveOdbDriver.InternalGetIndexName( const Handle : TOdbHandle ) : String;
begin
    Result := GetBtrieveFile(Handle).IndexName;
end;

{--------------------------------------------------------------------------}

procedure TBtrieveOdbDriver.InternalSetIndexName(
                      const Handle    : TOdbHandle;
                      const IndexName : String );
begin
    GetBtrieveFile(Handle).IndexName := IndexName;
end;

{--------------------------------------------------------------------------}

procedure TBtrieveOdbDriver.InternalGetCurrent(
                      const Handle  : TOdbHandle;
                      const Stream  : TMemoryStream );
var
    BtrieveFile : TBtrieveFile;
begin
    BtrieveFile := GetBtrieveFile(Handle);

    BtrieveFile.ReadObjectData(Stream);
end;

{--------------------------------------------------------------------------}

function TBtrieveOdbDriver.InternalLockCurrent( const Handle : TOdbHandle ) : Boolean;
begin
    Result := GetBtrieveFile(Handle).LockCurrent;
end;

{--------------------------------------------------------------------------}

procedure TBtrieveOdbDriver.InternalUnLockCurrent( const Handle : TOdbHandle );
begin
    GetBtrieveFile(Handle).UnLockCurrent;
end;

{--------------------------------------------------------------------------}

function TBtrieveOdbDriver.InternalFindNext( const Handle : TOdbHandle ) : Boolean;
begin
    Result := GetBtrieveFile(Handle).Next;
end;

{--------------------------------------------------------------------------}

function TBtrieveOdbDriver.InternalFindPrior( const Handle : TOdbHandle ) : Boolean;
begin
    Result := GetBtrieveFile(Handle).Prior;
end;

{--------------------------------------------------------------------------}

function TBtrieveOdbDriver.InternalFindFirst( const Handle : TOdbHandle ) : Boolean;
begin
    Result := GetBtrieveFile(Handle).First;
end;

{--------------------------------------------------------------------------}

function TBtrieveOdbDriver.InternalFindLast( const Handle : TOdbHandle ) : Boolean;
begin
    Result := GetBtrieveFile(Handle).Last;
end;

{--------------------------------------------------------------------------}

function TBtrieveOdbDriver.InternalFindExact(
                    const Handle : TOdbHandle;
                    const Fields : TUnOrderedList;
                    const Index  : TOdbIndex ) : Boolean;
begin
    Result := GetBtrieveFile(Handle).FindExact(Fields);
end;

{--------------------------------------------------------------------------}

procedure TBtrieveOdbDriver.InternalFindNearest(
                      const Handle : TOdbHandle;
                      const Fields : TUnOrderedList;
                      const Index  : TOdbIndex );
begin
    GetBtrieveFile(Handle).FindNearest(Fields);
end;

{--------------------------------------------------------------------------}

function TBtrieveOdbDriver.InternalGetBookmark( const Handle : TOdbHandle ) : Pointer;
begin
    Result := GetBtrieveFile(Handle).GetBookmark;
end;

{--------------------------------------------------------------------------}

procedure TBtrieveOdbDriver.InternalGotoBookmark( const Handle   : TOdbHandle;
                                                  const Bookmark : Pointer );
begin
    GetBtrieveFile(Handle).GotoBookmark(Bookmark);
end;

{--------------------------------------------------------------------------}

procedure TBtrieveOdbDriver.InternalFreeBookmark( const Handle   : TOdbHandle;
                                              const Bookmark : Pointer );
begin
    GetBtrieveFile(Handle).FreeBookmark(Bookmark);
end;

{--------------------------------------------------------------------------}

function TBtrieveOdbDriver.InternalGetObjectCount( const Handle : TOdbHandle ) : LongInt;
begin
    Result := GetBtrieveFile(Handle).RecordCount;
end;

{--------------------------------------------------------------------------}

function TBtrieveOdbDriver.GetBtrieveFile( const Handle : TOdbHandle ) : TBtrieveFile;
begin
    Result := GetHandleDataObj(Handle) as TBtrieveFile;
end;

{--------------------------------------------------------------------------}

function TBtrieveOdbDriver.GetSupportedFieldTypes : TOdbFieldTypes;
begin
    Result := [oftString, oftInteger, oftFloat, oftAutoInc];
end;

{--------------------------------------------------------------------}

function TBtrieveOdbDriver.GetDataAccessMethods : TDataAccessMethodSet;
begin
    Result := [damISAM];
end;

{--------------------------------------------------------------------------}

function TBtrieveOdbDriver.AutoIncSupported              : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TBtrieveOdbDriver.IntegerIndexSupported         : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TBtrieveOdbDriver.FloatIndexSupported           : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TBtrieveOdbDriver.CaseInsensitiveIndexSupported : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TBtrieveOdbDriver.DescendingIndexSupported      : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TBtrieveOdbDriver.TransactionsSupported         : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TBtrieveOdbDriver.FileLockingSupported          : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TBtrieveOdbDriver.ObjectLockingSupported        : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TBtrieveOdbDriver.SingleIndexOnly               : Boolean;
begin
    Result := False;
end;

{--------------------------------------------------------------------}

{$ifdef __Testing}

procedure TBtrieveOdbDriverTest.GetParams(
                    var DriverClass                   : TOdbDriverClass;
                    var DatabaseName                  : String;
                    var ServerName                    : String;
                    var UserName                      : String;
                    var Password                      : String
                  );
begin
    DriverClass  := TBtrieveOdbDriver;
    DatabaseName := Self.TestDbPath;
    ServerName   := '';
    UserName     := '';
    Password     := '';
end;

{$endif}

{--------------------------------------------------------------------------}

begin
    RegisterOdbDriver(TBtrieveOdbDriver, 'Btrieve');
end.
