{----------------------------------------------------------------------------
|
| Library: Spider Object Database
|
| Module: OdbQdb.Pas
|
| Description: QDB object database driver class. Tested with QDB version 2.01.
|
| History: Jan 17, 1997. Michel Brazeau, first version
|          Feb 19, 2001. MB override introspection methods and testcase
|
|---------------------------------------------------------------------------}
unit OdbQdb;

{$I Odb.Inc}

interface

uses
    {$ifdef __Testing}
    DriverTest, { for TOdbDriverTest }
    {$endif}


    SysUtils, { for IntToStr, Exception }
    Classes,  { for TMemoryStream }
    ObjList,  { for TUnOrderedList }
    Qdb,      { for TQdb }
    Odb,      { for TObjectFile, TOdbDriver }
    OdbIndex, { for TOdbIndex3 }
    OdbField; { for TOdbField }

type

TQdbFile = class(TQdb)
public
    constructor Create( AOwner : TComponent ); override;

protected
    { The FileName property of the TQdb file may be blank when it is closed.
      For this reason, the DiskFileName is stored separately. }
    DiskFileName : String;

    IndexName    : String;

    procedure BeforeKillHandler( Sender: TObject; var OK: Boolean );

    { event used to fix bug to prevent the EQdbNoData exception from being
      raised when the last object is deleted. }
    procedure OnNoData( Sender : TObject );
end;

{--------------------------------------------------------------------------}

{ abstract object database driver }
TQdbOdbDriver = class(TOdbDriver)
private
    FAutoSave : Boolean;

public

    constructor Create(const InDatabaseName : String); override;
    destructor Destroy; override;

    function InternalOpenDataset( const ObjectDataset : TObjectDataset ) : TOdbHandle; override;

    procedure InternalCloseDataset( const Handle : TOdbHandle); override;

    procedure InternalDeleteDataset( const ObjectDataset : TObjectDataset ); override;

    procedure InternalCreateDataset( const ObjectDataset : TObjectDataset ); override;

    procedure InternalInsert(
                      const Handle  : TOdbHandle;
                      const Fields  : TUnOrderedList;
                      const Stream  : TMemoryStream ); override;

    procedure InternalUpdate(
                      const Handle     : TOdbHandle;
                      const Fields     : TUnOrderedList;
                      const Stream     : TMemoryStream ); override;

    procedure InternalDelete( const Handle : TOdbHandle ); override;

    function InternalGetIndexName( const Handle : TOdbHandle ) : String; override;

    procedure InternalSetIndexName(
                            const Handle    : TOdbHandle;
                            const IndexName : String ); override;

    procedure InternalGetCurrent(
                          const Handle  : TOdbHandle;
                          const Stream  : TMemoryStream ); override;

    function InternalFindNext( const Handle : TOdbHandle ) : Boolean; override;

    function InternalFindPrior( const Handle : TOdbHandle ) : Boolean; override;

    function InternalFindFirst( const Handle : TOdbHandle ) : Boolean; override;

    function InternalFindLast( const Handle : TOdbHandle ) : Boolean; override;

    function InternalFindExact(
                        const Handle : TOdbHandle;
                        const Fields : TUnOrderedList;
                        const Index  : TOdbIndex ) : Boolean; override;

    procedure InternalFindNearest(
                          const Handle : TOdbHandle;
                          const Fields : TUnOrderedList;
                          const Index  : TOdbIndex ); override;

    function InternalGetBookmark( const Handle : TOdbHandle ) : Pointer; override;
    procedure InternalGotoBookmark( const Handle   : TOdbHandle;
                                    const Bookmark : Pointer ); override;
    procedure InternalFreeBookmark( const Handle   : TOdbHandle;
                                    const Bookmark : Pointer ); override;

    function InternalGetObjectCount( const Handle : TOdbHandle ) : LongInt; override;

    procedure InternalPack( const Handle : TOdbHandle ); override;

    function GetSupportedFieldTypes : TOdbFieldTypes; override;

    function CheckDuplicateUpdate : Boolean; override;

    function AutoIncSupported              : Boolean; override;
    function IntegerIndexSupported         : Boolean; override;
    function FloatIndexSupported           : Boolean; override;
    function CaseInsensitiveIndexSupported : Boolean; override;
    function DescendingIndexSupported      : Boolean; override;
    function TransactionsSupported         : Boolean; override;
    function FileLockingSupported          : Boolean; override;
    function ObjectLockingSupported        : Boolean; override;
    function SingleIndexOnly               : Boolean; override;


    { Qdb files uses an in memory index while storing data externally. This
      may lead to integrity problems on system failures. Setting AutoSave to
      True will ensure the index is written to disk after every Insert, Update,
      or Delete operation. This improves data integrity at the expense of a
      relatively large performance hit. By Default AutoSave is True. }
    property AutoSave : Boolean read FAutoSave
                                write FAutoSave;

    { Returns the underlying Qdb file corresponding to an object file handle.
      This may be used to invoke the Save method of the Qdb file explicitely
      if the AutoSave property is set False to improve performance. }
    function GetQdbFile( const Handle : TOdbHandle ) : TQdbFile;

private

    procedure ValidateFileStructure( const Fields  : TUnOrderedList;
                                     const Indices : TUnOrderedList );

end;

{--------------------------------------------------------------------------}

EQdbIndexMustBeUnique        = class(Exception);
EQdbFileMustHaveOnlyOneIndex = class(Exception);
EQdbFileMustHaveOnlyOneField = class(Exception);


{$ifdef __Testing}

type

TQdbOdbDriverTest = class( TOdbDriverTest )
protected
     procedure GetParams( var DriverClass                   : TOdbDriverClass;
                          var DatabaseName                  : String;
                          var ServerName                    : String;
                          var UserName                      : String;
                          var Password                      : String
                        ); override;
end;

{$endif}

{--------------------------------------------------------------------}

implementation

uses
    OdbMisc,   { for GetLocalObjectFilename }
    FileCtrl;  { for DirectoryExists }

{--------------------------------------------------------------------------}

const

CQdbFileExtension = '.Qdb';

{--------------------------------------------------------------------------}

constructor TQdbFile.Create( AOwner : TComponent );
begin
    inherited Create(AOwner);
    BeforeKill := BeforeKillHandler;
    WarnNoData := OnNoData;
end;

{--------------------------------------------------------------------------}

procedure TQdbFile.BeforeKillHandler( Sender: TObject; var OK: Boolean );
begin
    Ok := True;
end;

{--------------------------------------------------------------------------}

procedure TQdbFile.OnNoData( Sender : TObject );
begin
end;

{--------------------------------------------------------------------------}

constructor TQdbOdbDriver.Create(const InDatabaseName : String);
begin
    if (InDatabaseName <> '') and
       not DirectoryExists(InDatabaseName) then
        RaiseOdbException( EOdbDatabaseDoesNotExist, InDatabaseName);

    inherited Create(InDatabaseName);

    FAutoSave := True;
end;

{--------------------------------------------------------------------------}

destructor TQdbOdbDriver.Destroy;
begin
    inherited Destroy;
end;

{--------------------------------------------------------------------------}

procedure TQdbOdbDriver.ValidateFileStructure(
                                     const Fields  : TUnOrderedList;
                                     const Indices : TUnOrderedList );
var
    Index     : TOdbIndex;
begin
    { perform validation for Qdb files. Only one field of type oftString
      is allowed, and only one index. }
    if Fields.Size <> 1 then
        RaiseOdbException( EQdbFileMustHaveOnlyOneField, '' );

    if Indices.Size <> 1 then
        RaiseOdbException( EQdbFileMustHaveOnlyOneIndex, '' );

    Indices.GotoFirst;
    Index := (Indices.CurrentObj as TOdbIndex);
    if (not (iaUnique in Index.Attributes)) and
       (not (iaPrimary in Index.Attributes)) then
        RaiseOdbException( EQdbIndexMustBeUnique, '' );
end;

{--------------------------------------------------------------------------}

function TQdbOdbDriver.GetQdbFile( const Handle : TOdbHandle ) : TQdbFile;
begin
    Result := GetHandleDataObj(Handle) as TQdbFile;
end;

{--------------------------------------------------------------------------}

function TQdbOdbDriver.InternalOpenDataset(
                            const ObjectDataset : TObjectDataset ) : TOdbHandle;
var
    QdbFile    : TQdbFile;
    Index      : TOdbIndex;
    ObjectFile : TObjectFile;
begin
    ObjectFile := ObjectDataset as TObjectFile;

    ValidateFileStructure(ObjectDataset.Fields, ObjectFile.Indices);

    QdbFile := TQdbFile.Create(nil);
    try
        QdbFile.DiskFileName := GetLocalObjectFilename( DatabaseName, ObjectDataset.DatasetName,
                                                        CQdbFileExtension );
        if not FileExists(QdbFile.DiskFileName) then
            RaiseOdbException( EOdbObjectFileDoesNotExist, QdbFile.DiskFileName );

        QdbFile.FileName     := QdbFile.DiskFileName;

        ObjectFile.Indices.GotoFirst;
        Index := ObjectFile.Indices.CurrentObj as TOdbIndex;
        QdbFile.IndexName := Index.Name;
        QdbFile.KeyCaseSensitive := not (iaCaseInsensitive in Index.Attributes);

        Result := NewHandle(QdbFile);
    except
        QdbFile.Free;
        raise;
    end;
end;

{--------------------------------------------------------------------------}

procedure TQdbOdbDriver.InternalCloseDataset( const Handle : TOdbHandle);
begin
    GetQdbFile(Handle).FileName := '';
    ReleaseHandle(Handle);
end;

{--------------------------------------------------------------------------}

procedure TQdbOdbDriver.InternalDeleteDataset( const ObjectDataset : TObjectDataset );
var
    QdbFile : TQdbFile;
begin
    QdbFile := TQdbFile.Create(nil);
    try
        QdbFile.DiskFileName := GetLocalObjectFileName( DatabaseName,
                                                        ObjectDataset.DatasetName,
                                                        CQdbFileExtension );
        QdbFile.FileName := QdbFile.DiskFileName;
        QdbFile.Kill;
    finally
        QdbFile.Free;
    end;
end;

{--------------------------------------------------------------------------}

procedure TQdbOdbDriver.InternalCreateDataset(
                          const ObjectDataset : TObjectDataset
                        );
var
    QdbFile      : TQdbFile;
    DiskFileName : String;
    ObjectFile   : TObjectFile;
begin
    ObjectFile := ObjectDataset as TObjectFile;

    ValidateFileStructure( ObjectFile.Fields, ObjectFile.Indices );

    DiskFileName := GetLocalObjectFileName( DatabaseName,
                                            ObjectDataset.DatasetName,
                                            CQdbFileExtension );

    if FileExists(DiskFileName) then
        Self.DeleteDataset(ObjectFile);

    QdbFile := TQdbFile.Create(nil);
    try
        QdbFile.FileName     := DiskFileName;
    finally
        QdbFile.Free;
    end;
end;

{--------------------------------------------------------------------------}

procedure TQdbOdbDriver.InternalInsert(
                  const Handle  : TOdbHandle;
                  const Fields  : TUnOrderedList;
                  const Stream  : TMemoryStream );
var
    QdbFile     : TQdbFile;
begin
    QdbFile := GetQdbFile(Handle);
    Fields.GotoFirst;
    QdbFile.Add(Stream, (Fields.CurrentObj as TOdbField).AsString);

    if AutoSave then
        QdbFile.Save;
end;

{--------------------------------------------------------------------------}

procedure TQdbOdbDriver.InternalUpdate(
                  const Handle     : TOdbHandle;
                  const Fields     : TUnOrderedList;
                  const Stream     : TMemoryStream );
var
    QdbFile : TQdbFile;
begin
    QdbFile := GetQdbFile(Handle);

    QdbFile.Delete;
    Fields.GotoFirst;
    QdbFile.Add(Stream, (Fields.CurrentObj as TOdbField).AsString);

    if AutoSave then
        QdbFile.Save;
end;

{--------------------------------------------------------------------------}

procedure TQdbOdbDriver.InternalDelete( const Handle : TOdbHandle );
var
    QdbFile   : TQdbFile;
    FirstItem : Boolean;
begin
    QdbFile   := GetQdbFile(Handle);
    FirstItem := (QdbFile.ItemIndex = 0);
    QdbFile.Delete;

    if AutoSave then
        QdbFile.Save;

    { the Qdb default behavior is to move to the previous item, but the
      TObjectFile's behavior is to move to the next item.

      MB Feb 28, 2001. When deleting the first item, QDB loses position and
      a next moves to the second item. Only move to the next item, if it was
      not the first item being deleted. }
    if FirstItem then
        QdbFile.FirstItem
    else
        QdbFile.NextItem;
end;

{--------------------------------------------------------------------------}

function TQdbOdbDriver.InternalGetIndexName( const Handle : TOdbHandle ) : String;
begin
    Result := GetQdbFile(Handle).IndexName;
end;

{--------------------------------------------------------------------------}

procedure TQdbOdbDriver.InternalSetIndexName(
                      const Handle    : TOdbHandle;
                      const IndexName : String );
begin
    if GetQdbFile(Handle).IndexName <> IndexName then
        RaiseOdbException( EOdbIndexDoesNotExist, IndexName );
end;

{--------------------------------------------------------------------------}

procedure TQdbOdbDriver.InternalGetCurrent(
                      const Handle  : TOdbHandle;
                      const Stream  : TMemoryStream );
begin
    GetQdbFile(Handle).Get(Stream);
end;

{--------------------------------------------------------------------------}

function TQdbOdbDriver.InternalFindNext( const Handle : TOdbHandle ) : Boolean;
var
    QdbFile    : TQdbFile;
    WasEOF     : Boolean;
begin
    QdbFile := GetQdbFile(Handle);

    WasEOF := QdbFile.EOF;

    QdbFile.NextItem;

    { MB Feb 20, 2001. Check if EOF was previous set }
    Result := (not QdbFile.EOF) or
              (QdbFile.EOF and not WasEOF);
end;

{--------------------------------------------------------------------------}

function TQdbOdbDriver.InternalFindPrior( const Handle : TOdbHandle ) : Boolean;
var
    QdbFile : TQdbFile;
    WasBOF     : Boolean;
begin
    QdbFile := GetQdbFile(Handle);

    WasBOF  := QdbFile.BOF;

    QdbFile.PrevItem;

    { MB Feb 20, 2001. Check if BOF was previous set }
    Result := (not QdbFile.BOF) or
             (QdbFile.BOF and not WasBOF);
end;

{--------------------------------------------------------------------------}

function TQdbOdbDriver.InternalFindFirst( const Handle : TOdbHandle ) : Boolean;
var
    QdbFile : TQdbFile;
begin
    QdbFile := GetQdbFile(Handle);

    { MB Feb 28, 2000. Simplify code, check BOF is adequate }
    QdbFile.FirstItem;

    Result := QdbFile.BOF;
end;

{--------------------------------------------------------------------------}

function TQdbOdbDriver.InternalFindLast( const Handle : TOdbHandle ) : Boolean;
var
    QdbFile : TQdbFile;
begin
    QdbFile := GetQdbFile(Handle);

    { MB Feb 28, 2000. Simplify code, check EOF is adequate }
    QdbFile.LastItem;

    Result := QdbFile.EOF;
end;

{--------------------------------------------------------------------------}

function TQdbOdbDriver.InternalFindExact(
                    const Handle : TOdbHandle;
                    const Fields : TUnOrderedList;
                    const Index  : TOdbIndex ) : Boolean;
var
    QdbFile : TQdbFile;
    Field   : TOdbField;
begin
    QdbFile := GetQdbFile(Handle);
    Fields.GotoFirst;
    Field   := Fields.CurrentObj as TOdbField;
    Result := QdbFile.ExactMatch(Field.AsString);
end;

{--------------------------------------------------------------------------}

procedure TQdbOdbDriver.InternalFindNearest(
                      const Handle : TOdbHandle;
                      const Fields : TUnOrderedList;
                      const Index  : TOdbIndex );
var
    QdbFile      : TQdbFile;
    Field        : TOdbField;
    Low          : LongInt;
    High         : LongInt;
    Pos          : LongInt;
    CompareValue : Integer;
begin
    QdbFile := GetQdbFile(Handle);

    Fields.GotoFirst;
    Field  := Fields.CurrentObj as TOdbField;

    { MB Feb 26, 2001. QDB does not have "greater than or equal" search.
      CloseMatch only works on partial strings, ie, Object4 would find
      Object41, but Object35 does not find Object4. This requires
      performing a binary search. Unfortunately the Qdb.Index property is
      private. }
    Low   := 0;
    High  := QdbFile.Count - 1;
    while Low <= High do
    begin
        Pos := (Low + High) div 2;
        QdbFile.ItemIndex := Pos;
        if QdbFile.KeyCaseSensitive then
            CompareValue := AnsiCompareStr(QdbFile.Key, Field.AsString)
        else
            CompareValue := AnsiCompareText(QdbFile.Key, Field.AsString);

        if CompareValue < 0 then
            Low := Pos + 1
        else
        begin
            High := Pos - 1;
            if CompareValue = 0 then
            begin
                Low    := Pos;
            end;
        end;
    end;

    if Low >= QdbFile.Count then
        QdbFile.ItemIndex := Low - 1
    else
        QdbFile.ItemIndex := Low;

    { By bypassing the QDB search methods, the EOF, BOF properties are not
      set }
    if QdbFile.ItemIndex = 0 then
        QdbFile.FirstItem;

    if QdbFile.ItemIndex = (QdbFile.Count-1) then
        QdbFile.LastItem;
end;

{--------------------------------------------------------------------------}

function TQdbOdbDriver.InternalGetBookmark( const Handle : TOdbHandle ) : Pointer;
var
    QdbFile    : TQdbFile;
    pItemIndex : ^LongInt;
begin
    QdbFile := GetQdbFile(Handle);

    GetMem(pItemIndex, SizeOf(LongInt));
    pItemIndex^ := QdbFile.ItemIndex;
    Result      := pItemIndex;
end;

{--------------------------------------------------------------------------}

procedure TQdbOdbDriver.InternalGotoBookmark( const Handle   : TOdbHandle;
                                              const Bookmark : Pointer );
var
    QdbFile    : TQdbFile;
    pItemIndex : ^LongInt;
begin
    QdbFile           := GetQdbFile(Handle);
    pItemIndex        := Bookmark;
    QdbFile.ItemIndex := pItemIndex^;
end;

{--------------------------------------------------------------------------}

procedure TQdbOdbDriver.InternalFreeBookmark( const Handle   : TOdbHandle;
                                              const Bookmark : Pointer );
begin
    FreeMem(Bookmark, SizeOf(LongInt));
end;


{--------------------------------------------------------------------------}

function TQdbOdbDriver.InternalGetObjectCount( const Handle : TOdbHandle ) : LongInt;
begin
    Result := GetQdbFile(Handle).Count;
end;

{--------------------------------------------------------------------------}

procedure TQdbOdbDriver.InternalPack( const Handle : TOdbHandle );
begin
    GetQdbFile(Handle).Pack;
end;

{--------------------------------------------------------------------------}

function TQdbOdbDriver.GetSupportedFieldTypes : TOdbFieldTypes;
begin
    Result := [oftString];
end;

{--------------------------------------------------------------------------}

function TQdbOdbDriver.CheckDuplicateUpdate : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TQdbOdbDriver.AutoIncSupported              : Boolean;
begin
    Result := False;
end;

{--------------------------------------------------------------------------}

function TQdbOdbDriver.IntegerIndexSupported         : Boolean;
begin
    Result := False;
end;

{--------------------------------------------------------------------------}

function TQdbOdbDriver.FloatIndexSupported           : Boolean;
begin
    Result := False;
end;

{--------------------------------------------------------------------------}

function TQdbOdbDriver.CaseInsensitiveIndexSupported : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TQdbOdbDriver.DescendingIndexSupported      : Boolean;
begin
    Result := False;
end;

{--------------------------------------------------------------------------}

function TQdbOdbDriver.TransactionsSupported         : Boolean;
begin
    Result := False;
end;

{--------------------------------------------------------------------------}

function TQdbOdbDriver.FileLockingSupported          : Boolean;
begin
    Result := False;
end;

{--------------------------------------------------------------------------}

function TQdbOdbDriver.ObjectLockingSupported        : Boolean;
begin
    Result := False;
end;

{--------------------------------------------------------------------------}

function TQdbOdbDriver.SingleIndexOnly               : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}


{$ifdef __Testing}

procedure TQdbOdbDriverTest.GetParams(
                    var DriverClass                   : TOdbDriverClass;
                    var DatabaseName                  : String;
                    var ServerName                    : String;
                    var UserName                      : String;
                    var Password                      : String
                  );
begin
    DriverClass  := TQdbOdbDriver;
    DatabaseName := Self.TestDbPath;
    ServerName   := '';
    UserName     := '';
    Password     := '';
end;

{$endif}


{--------------------------------------------------------------------------}

begin
    RegisterOdbDriver(TQdbOdbDriver, 'QDB');
end.
