{----------------------------------------------------------------------
|
| Library: Spider Object Database
|
| Module: TestObj.Pas
|
| Description: Classes used with automated tests for Spider ODB
|              class.
|
| History: Feb 03, 2001. Michel Brazeau, first version
|
|---------------------------------------------------------------------}
unit TestObj;

{$I Odb.Inc}

{$ifndef __Testing}
    !! This unit is for automated tests only
{$endif}

interface

uses
    AutoTest, { for TSpiderOdbTestcase }
    Classes,  { for TMemoryStream }
    Odb;      { for TObjectFile, TObjectQuery }

{--------------------------------------------------------------------------}

type

TTestObject = class(TObject)

    Name     : String;
    Age      : LongInt;
    Id       : LongInt;
    Weight   : Double;

    KeyValue : String;

    Data     : LongInt;

    Stream   : TMemoryStream;

    constructor Create;

    destructor Destroy; override;
end;

{--------------------------------------------------------------------------}

TTestObjectMapping = class(TORMapping)
public
    { List of methods fired }
    FiredMethods  : TStringList;

    constructor Create( const InDatabase : TObjectDatabase ); override;
    destructor Destroy; override;

    function GetObjectType : TClass; override;

    procedure DefineFields; override;
    procedure DefineIndices; override;
    procedure UpdateFieldsFromObj( const Obj       : TObject ); override;

    { This method is overriden when the table has an autoincrement field }
    procedure UpdateObjFromFields( const Obj       : TObject ); override;

    procedure BeforeObjectPost( const Stream  : TMemoryStream ); override;
    procedure AfterObjectRead( const Stream  : TMemoryStream ); override;

    procedure BeforeInsert; override;
    procedure AfterInsert; override;

    procedure BeforeUpdate; override;
    procedure AfterUpdate; override;

    procedure BeforeDelete; override;
    procedure AfterDelete; override;

end;

{--------------------------------------------------------------------------}

{ A class is derived from TObjectFile to test the Open/Close hooks. Usually
  it is not necessary to derive from TObjectFile when the TORMapping class
  is written }
TTestObjectFile = class(TObjectFile)

    procedure BeforeOpen; override;
    procedure AfterOpen; override;

    procedure BeforeClose; override;
    procedure AfterClose; override;
end;

{ routines to add objects for test purposes }
procedure AddObjectISAM( const Age        : LongInt;
                         const ObjectFile : TObjectFile );
procedure AddObjectSQL( const Age          : LongInt;
                        const ObjectUpdate : TObjectUpdate );


{ check the values of an object added with AddObjectISAM or AddObjectSQL }
procedure CheckObjectValues( const Age        : LongInt;
                             const TestObject : TTestObject );


(************************************************************************

  * *   * ***  *    **** *   * **** *   * *****   *   ***** *  **  *   *
  * ** ** *  * *    *    ** ** *    **  *   *    * *    *   * *  * **  *
  * * * * ***  *    **   * * * **   * * *   *   *   *   *   * *  * * * *
  * *   * *    *    *    *   * *    *  **   *   *****   *   * *  * *  **
  * *   * *    **** **** *   * **** *   *   *   *   *   *   *  **  *   *

 ************************************************************************)

implementation

uses
    SysUtils,  { for Exception }
    ObjPerst,  { for TObjectWriter, TObjectReader }
    OdbCompr,  { for CompressStreamRLE }
    OdbIndex,  { for iaXXX }
    OdbField;  { for oftXXX }

{--------------------------------------------------------------------------}

constructor TTestObject.Create;
var
    Byte1, Byte2, Byte3, Byte4 : Byte;
begin
    inherited Create;

    Stream  := TMemoryStream.Create;
    Stream.SetSize(1024);

    Byte1 := $01;
    Byte2 := $7F;
    Byte3 := $FE;
    Byte4 := $00;

    Stream.Seek(0, soFromBeginning);
    Stream.Write(Byte1, SizeOf(Byte1));

    Stream.Seek(511, soFromBeginning);
    Stream.Write(Byte2, SizeOf(Byte2));

    Stream.Seek(1023, soFromBeginning);
    Stream.Write(Byte3, SizeOf(Byte3));

    Stream.Seek(555, soFromBeginning);
    Stream.Write(Byte4, SizeOf(Byte4));
end;

{--------------------------------------------------------------------------}

destructor TTestObject.Destroy;
begin
    Stream.Free;

    inherited Destroy;
end;

{--------------------------------------------------------------------------}

procedure ReadTestObject( const Obj     : TObject;
                          const Reader  : TObjectReader ); far;
var
    TestObject   : TTestObject;
    Byte1        : Byte;
    Byte2        : Byte;
    Byte3        : Byte;
    Byte4        : Byte;
begin
    TestObject          := Obj as TTestObject;
    TestObject.Name     := Reader.ReadString;
    TestObject.KeyValue := Reader.ReadString;
    TestObject.Age      := Reader.ReadInteger;
    TestObject.Id       := Reader.ReadInteger;
    TestObject.Weight   := Reader.ReadFloat;
    TestObject.Data     := Reader.ReadInteger;
    TestObject.Stream   := Reader.ReadChildObject as TMemoryStream;

    TestObject.Stream.Seek(0, soFromBeginning);
    TestObject.Stream.Read(Byte1, SizeOf(Byte1));

    TestObject.Stream.Seek(511, soFromBeginning);
    TestObject.Stream.Read(Byte2, SizeOf(Byte2));

    TestObject.Stream.Seek(1023, soFromBeginning);
    TestObject.Stream.Read(Byte3, SizeOf(Byte3));

    TestObject.Stream.Seek(555, soFromBeginning);
    TestObject.Stream.Read(Byte4, SizeOf(Byte4));

    if (Byte1 <> $01) or
       (Byte2 <> $7F) or
       (Byte3 <> $FE) or
       (Byte4 <> $00) then
        raise Exception.Create('TestObject.Stream is corrupt');
end;

{--------------------------------------------------------------------------}

procedure WriteTestObject( const Obj     : TObject;
                           const Writer  : TObjectWriter ); far;
var
    TestObject : TTestObject;
begin
    TestObject := Obj as TTestObject;
    Writer.WriteString(TestObject.Name);
    Writer.WriteString(TestObject.KeyValue);
    Writer.WriteInteger(TestObject.Age);
    Writer.WriteInteger(TestObject.Id);
    Writer.WriteFloat(TestObject.Weight);
    Writer.WriteInteger(TestObject.Data);

    Writer.WriteChildObject(TestObject.Stream);
end;

{--------------------------------------------------------------------------}

constructor TTestObjectMapping.Create( const InDatabase : TObjectDatabase );
begin
    inherited Create(InDatabase);

    FiredMethods := TStringList.Create;
end;

{--------------------------------------------------------------------------}

destructor TTestObjectMapping.Destroy;
begin
    FiredMethods.Free;

    inherited Destroy;
end;

{--------------------------------------------------------------------------}

function TTestObjectMapping.GetObjectType : TClass;
begin
    Result := TTestObject;
end;

{--------------------------------------------------------------------------}

procedure TTestObjectMapping.UpdateFieldsFromObj( const Obj     : TObject );
begin
    FieldByName('Name').AsString := (Obj as TTestObject).Name;

    if not FDatabase.Driver.SingleIndexOnly then
        FieldByName('KeyValue').AsString := (Obj as TTestObject).KeyValue;

    if FDatabase.Driver.IntegerIndexSupported then
        FieldByName('Age').AsInteger := (Obj as TTestObject).Age;

    if FDatabase.Driver.FloatIndexSupported then
        FieldByName('Weight').AsFloat := (Obj as TTestObject).Weight;

    if FDatabase.Driver.AutoIncSupported then
        FieldByName('Id').AsInteger := (Obj as TTestObject).Id;
end;

{--------------------------------------------------------------------------}

procedure TTestObjectMapping.DefineFields;
begin
    AddField('Name', oftString, 41);

    if not FDatabase.Driver.SingleIndexOnly then
        AddField('KeyValue', oftString, 21);

    if FDatabase.Driver.IntegerIndexSupported then
        AddField('Age', oftInteger, 0);

    if FDatabase.Driver.FloatIndexSupported then
        AddField('Weight', oftFloat, 0);

    if FDatabase.Driver.AutoIncSupported then
        AddField('Id', oftAutoInc, 0);
end;

{--------------------------------------------------------------------------}

procedure TTestObjectMapping.DefineIndices;
begin
    AddIndex( 'Name', [iaPrimary], ['Name']);

    if FDatabase.Driver.SingleIndexOnly then
        Exit;

    AddIndex( 'KeyValue', [iaUnique], ['KeyValue']);

    if FDatabase.Driver.CaseInsensitiveIndexSupported then
        AddIndex( 'NameCI', [iaUnique, iaCaseInsensitive], ['Name']);

    if FDatabase.Driver.IntegerIndexSupported then
    begin
        AddIndex( 'Age', [], ['Age']);
        AddIndex( 'AgeName', [], ['Age','Name']);

        if FDatabase.Driver.DescendingIndexSupported then
            AddIndex( 'AgeDesc', [iaDescending], ['Age']);
    end;

    if FDatabase.Driver.FloatIndexSupported then
    begin
        AddIndex( 'Weight', [], ['Weight']);
        AddIndex( 'WName', [], ['Weight','Name'] );
    end;

    if FDatabase.Driver.FloatIndexSupported and
       FDatabase.Driver.IntegerIndexSupported then
        AddIndex( 'WAge', [], ['Weight','Age'] );

    if FDatabase.Driver.AutoIncSupported then
        AddIndex( 'Id', [iaUnique], ['Id']);
end;

{--------------------------------------------------------------------------}

procedure TTestObjectMapping.UpdateObjFromFields( const Obj     : TObject );
begin
    if FDatabase.Driver.AutoIncSupported then
        (Obj as TTestObject).Id := FieldByName('Id').AsInteger;
end;

{--------------------------------------------------------------------------}

procedure TTestObjectMapping.BeforeObjectPost( const Stream  : TMemoryStream );
begin
    if FiredMethods.IndexOf('BeforeObjectPost') = -1 then
        FiredMethods.Add('BeforeObjectPost');

   CompressStreamRLE(Stream);
end;

{--------------------------------------------------------------------------}

procedure TTestObjectMapping.AfterObjectRead( const Stream  : TMemoryStream );
begin
    if FiredMethods.IndexOf('AfterObjectRead') = -1 then
        FiredMethods.Add('AfterObjectRead');

   UnCompressStreamRLE(Stream);
end;

{--------------------------------------------------------------------------}

procedure TTestObjectMapping.BeforeInsert;
begin
    if FiredMethods.IndexOf('BeforeInsert') = -1 then
        FiredMethods.Add('BeforeInsert');
end;

{--------------------------------------------------------------------------}

procedure TTestObjectMapping.AfterInsert;
begin
    if FiredMethods.IndexOf('AfterInsert') = -1 then
        FiredMethods.Add('AfterInsert');
end;

{--------------------------------------------------------------------------}

procedure TTestObjectMapping.BeforeUpdate;
begin
    if FiredMethods.IndexOf('BeforeUpdate') = -1 then
        FiredMethods.Add('BeforeUpdate');
end;

{--------------------------------------------------------------------------}

procedure TTestObjectMapping.AfterUpdate;
begin
    if FiredMethods.IndexOf('AfterUpdate') = -1 then
        FiredMethods.Add('AfterUpdate');
end;

{--------------------------------------------------------------------------}

procedure TTestObjectMapping.BeforeDelete;
begin
    if FiredMethods.IndexOf('BeforeDelete') = -1 then
        FiredMethods.Add('BeforeDelete');
end;

{--------------------------------------------------------------------------}

procedure TTestObjectMapping.AfterDelete;
begin
    if FiredMethods.IndexOf('AfterDelete') = -1 then
        FiredMethods.Add('AfterDelete');
end;

{--------------------------------------------------------------------------}

procedure TTestObjectFile.BeforeOpen;
begin
    if (ORMapping as TTestObjectMapping).FiredMethods.IndexOf('BeforeOpen') = -1 then
        (ORMapping as TTestObjectMapping).FiredMethods.Add('BeforeOpen');
end;

{--------------------------------------------------------------------------}

procedure TTestObjectFile.AfterOpen;
begin
    if (ORMapping as TTestObjectMapping).FiredMethods.IndexOf('AfterOpen') = -1 then
        (ORMapping as TTestObjectMapping).FiredMethods.Add('AfterOpen');
end;

{--------------------------------------------------------------------------}

procedure TTestObjectFile.BeforeClose;
begin
    if (ORMapping as TTestObjectMapping).FiredMethods.IndexOf('BeforeClose') = -1 then
        (ORMapping as TTestObjectMapping).FiredMethods.Add('BeforeClose');
end;

{--------------------------------------------------------------------------}

procedure TTestObjectFile.AfterClose;
begin
    if (ORMapping as TTestObjectMapping).FiredMethods.IndexOf('AfterClose') = -1 then
        (ORMapping as TTestObjectMapping).FiredMethods.Add('AfterClose');
end;

{--------------------------------------------------------------------------}

procedure AddObjectISAM( const Age        : LongInt;
                         const ObjectFile : TObjectFile );
var
    TestObject : TTestObject;
begin
    TestObject := TTestObject.Create;
    try
        TestObject.Name     := 'Object' + IntToStr(Age);
        TestObject.KeyValue := IntToStr(Age);
        TestObject.Age      := Age;
        TestObject.Weight   := (-Age*10) - (Age*4) + (Age*0.45);

        ObjectFile.CurrentObj := TestObject;
        ObjectFile.Insert;
    finally
        TestObject.Free;
    end;
end;

{--------------------------------------------------------------------------}

procedure AddObjectSQL( const Age          : LongInt;
                        const ObjectUpdate : TObjectUpdate );
var
    TestObject : TTestObject;
begin
    TestObject := TTestObject.Create;
    try
        TestObject.Name      := 'Object' + IntToStr(Age);
        TestObject.KeyValue  := IntToStr(Age);
        TestObject.Age       := Age;
        TestObject.Weight    := (-Age*10) - (Age*4) + (Age*0.45);

        if not ObjectUpdate.Database.Driver.AutoincSupported then
            TestObject.Id        := Age;


        ObjectUpdate.Insert(TestObject);
    finally
        TestObject.Free;
    end;
end;

{--------------------------------------------------------------------------}

procedure CheckObjectValues( const Age        : LongInt;
                             const TestObject : TTestObject );
begin
    if TestObject.Name <> 'Object' + IntToStr(Age) then
        raise Exception.Create('CheckObjectValues: Name');

   if TestObject.Age <> Age then
        raise Exception.Create('CheckObjectValues: Age');

   if Round(TestObject.Weight*100) <> Round(((-Age*10) - (Age*4) + (Age*0.45))*100) then
        raise Exception.Create('CheckObjectValues: Weight');
end;

{--------------------------------------------------------------------------}

initialization

    RegisterStreamable( TTestObject,
                        [ReadTestObject],
                        [WriteTestObject]
                      );

end.
